
// Ranges and strides, and subscripts and indexes.

#include "RangeRep.h"
#include "CollapsedRangeRep.h"
#include "DimRangeRep.h"
#include "BlockRangeRep.h"
#include "CyclicRangeRep.h"
#include "StepRangeRep.h"

//#define MAX_LEVELS 2 

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Range': handle class for range representations.              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

//enum Layout {LAY_STANDARD, LAY_PACKED, LAY_ADAPT, LAY_EXTENDED} ;

class Stride ;
//class Location ;
class Block ;

#define DEF_LB  0
#define DEF_UB  -2

struct Trip {
  Trip() : lb(DEF_LB), ub(DEF_UB), str(1) {}
  Trip(const int _lb) : lb(_lb), ub(DEF_UB), str(1) {}
  Trip(const int _lb, const int _ub) : lb(_lb), ub(_ub), str(1) {}
  Trip(const int _lb, const int _ub, const int _str) :
      lb(_lb), ub(_ub), str(_str) {}

      // probably should check _ub >= -1.

  const int lb, ub, str ;
} ;

class Range {
public :
  // Constructors and destructors

  Range() : rep(0) {}

  Range(const Range& x) : rep(x.rep) {
    if(rep) rep->incr() ;
  }

  Range(const int extent) {
    rep = new CollapsedRangeRep(extent) ;
    rep->incr() ;
  }

  Range(Dimension dim) {
    rep = new DimRangeRep(dim) ;
    rep->incr() ;
  }

  ~Range() {if(rep) rep->decr() ; }

  // Subranges

  Range subrng(const int extent,
               const int base = 0, const int stride = 1) const {
    if(extent != 0) {
      const int parExtent = size() ;
      const int top = base + stride * (extent - 1) ;
      if(base < 0 || base >= parExtent || top < 0 || top >= parExtent)
        AdlibError("Range :: subrng: subrange out of range") ;
    }

    return Range(rep->subrng(extent, base, stride)) ;
  }

  Range triplet(const int base, const int top, const int stride = 1) const {
    int extent ;
    if((stride > 0 && top < base) || (stride < 0 && top > base))
      extent = 0 ;
    else {
      const int parExtent = size() ;
      if(base < 0 || base >= parExtent || top < 0 || top >= parExtent)
        AdlibError("Range :: trip: subrange out of range") ;
      else
        extent = (top - base) / stride + 1 ;
    }

    return Range(rep->subrng(extent, base, stride)) ;
  }

  // Inquiries

  int size() const {return rep->size() ; }

  Dimension dim() const {return rep->dim() ; }

  int lev() const {return rep->lev() ; }
  Range ker() const { return Range(rep->ker()) ; }
  Range shell() const { return Range(rep->shell()) ; }

  Range tem() const { return Range(rep->tem()) ; }
  int bas() const {return rep->bas() ; }
  int str() const {return rep->str() ; }

  Format format() const {return rep->format() ; }

  // Subscripting and iteration

  //int ker_glb(const int glb, int* sub) const {
  //  return rep->ker_glb(glb, sub) ;
  //}

  void location(Location* loc, const int glb) const {
    rep->location(loc, glb) ;
  }

  int block(const int ker_glb, int* sub, int* count) const {
    return rep->block(ker_glb, sub, count) ;
  }
  int block_step(int* sub_stp) const {
    return rep->block_step(sub_stp) ;
  }

  //int crd(const int glb) const {return rep->crd(glb) ; }
  int local(int* glb) const {return rep->local(glb) ; }

  Range subker(int* ker_bas, int* ker_str) const {
    return Range(rep->subker(ker_bas, ker_str)) ;
  }

  // `Stride' stuff

  int width() const {return rep->width() ; }

  int disp(const int sub) const {return rep->disp(sub) ; }
  int step(const int sub_stp) const {return rep->step(sub_stp) ; }

  int volume() const {return rep->volume() ; }

//  int offset(const int* subs) const {
//    if(lev())
//      return width() * ker().offset(subs + 1) + disp(*subs) ;
//    else
//      return 0 ;
//  }

  int offset(const Location& i) const {
    return rep->offset(&i) ;
  }

  Stride stride(int* size) const ;

  // Miscellaneous operators

  //Location operator()(const int k) ;

  Location operator()(const int glb) {
    Location res;
    location(&res, glb) ;
    return res ;
  }

  //int operator()(const Location& i) ;  // Must reinstate something equivalent.

  Range operator()(const Trip& t) {
    const int base   = t.lb ;
    const int top    = t.ub != DEF_UB ? t.ub : size() - 1 ;
    const int stride = t.str ;
    return triplet(base, top, stride) ;
  }

  Range& operator=(const Range& x) {
    if(x.rep) x.rep->incr() ;
    if(rep)   rep->decr() ;
    rep = x.rep ;
    return *this ;
  }

  operator RangeRep*() const {return rep ; }

  // Interim support for PCRC Fortran interface

  int glb(const int sub, const int ker_glb) const {
    return rep->glb(sub, ker_glb) ;
  }

  int& g_lb() {return rep->g_lb ; }
  int& g_stride() {return rep->g_stride ; }

  // debugging

  void show() {
    if(rep)
      rep->show() ;
    else
      AdlibPrintf("Range :: show: Null range\n") ;
  }
protected :
  Range(RangeRep* _rep) : rep(_rep) {
    if(rep) rep->incr() ;
  }

  RangeRep* rep ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockRange': block-distributed range.                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockRange : public Range {
public :
  BlockRange(const int extent, Dimension dim) {
    const int blockSize = (extent + dim.size() - 1) / dim.size() ;

    rep = new BlockRangeRep(extent, Range(dim), blockSize) ;
    rep->incr() ;
  }

  BlockRange(const int extent, Dimension dim, const int wlo, const int whi) {
    const int blockSize = (extent + dim.size() - 1) / dim.size() ;

    rep = new ExtendedBlockRangeRep(extent, Range(dim), blockSize, wlo, whi) ;
    rep->incr() ;
  }

  BlockRange(const int extent, Dimension dim,  const int blockSize) {
    rep = new BlockRangeRep(extent, Range(dim), blockSize) ;
    rep->incr() ;
  }

  BlockRange(const int extent, Dimension dim, const int blockSize,
	     const int wlo, const int whi) {
    rep = new ExtendedBlockRangeRep(extent, Range(dim), blockSize, wlo, whi) ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CyclicRange': cyclic-distributed range.                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class CyclicRange : public Range {
public :
  CyclicRange(const int extent, Dimension dim) {
    const int numCycles = (extent + dim.size() - 1) / dim.size() ;
    rep = new CyclicRangeRep(extent, Range(dim), numCycles) ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockCyclicRange': block-cyclically distributed range.       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockCyclicRange : public Range {
public :
  BlockCyclicRange(const int extent, Dimension dim, const int blockSize) {
    const int numBlocks = (extent + blockSize - 1) / blockSize ;
    const int numCycles = (numBlocks + dim.size() - 1) / dim.size() ;
    RangeRep* ker = new CyclicRangeRep(numBlocks, Range(dim), numCycles) ;
    rep = new BlockCyclicRangeRep(extent, ker, blockSize) ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `StepRange': stepped block-distributed range.                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class StepRange : public Range {
public :
  StepRange(const int extent, Dimension dim) {
    rep = new StepRangeRep(Range(dim), extent / dim.size(),
				       extent % dim.size()) ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ADAPTBlockRange': block-distributed range yielding arrays    //
//                           with special packed memory layout         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ADAPTBlockRange : public Range {
public :
  ADAPTBlockRange(const int extent, const int base, const int stride,
                  const int parExtent, Dimension dim, const int blockSize) {
    BlockRangeRep* par = new BlockRangeRep(parExtent, Range(dim), blockSize) ;

    if(extent != par->size() || stride < 0)
      rep = new PackedBlockSubRangeRep(extent, base, stride,
				       par, par->ker(), blockSize,
				       stride > 0 ? stride : -stride) ;
    else
      rep = par ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ADAPTCyclicRange': cyclic-distributed range yielding arrays  //
//                           with special truncated memory layout.     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ADAPTCyclicRange : public Range {
public :
  ADAPTCyclicRange(const int extent, const int base, const int stride,
                   Dimension dim) {
    const int parExtent =
        (stride > 0 ? base + stride * (extent - 1) : base) + 1 ;
    const int numCycles = (parExtent + dim.size() - 1) / dim.size() ;

    RangeRep* par = new CyclicRangeRep(parExtent, Range(dim), numCycles) ;

    rep = (extent != par->size() || stride < 0) ?
              par->subrng(extent, base, stride) : par ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ADAPTBlockCyclicRange': block-cyclically distributed range,  //
//                                yielding arrays with special packed  //
//                                memory layout.                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ADAPTBlockCyclicRange : public Range {
public :
  ADAPTBlockCyclicRange(const int extent, const int base, const int stride,
                        Dimension dim, const int blockSize) {
    const int numBlocks =
        (stride > 0 ? base + stride * (extent - 1) : base) / blockSize + 1 ;
    const int numCycles = (numBlocks + dim.size() - 1) / dim.size() ;

    RangeRep* ker = new CyclicRangeRep(numBlocks, Range(dim), numCycles) ;
    BlockRangeRep* par = new BlockCyclicRangeRep(numBlocks * blockSize,
                                                 ker, blockSize) ;

    if(extent != par->size() || stride < 0)
      rep = new PackedBlockCyclicSubRangeRep(extent, base, stride,
				             par, par->ker(), blockSize,
				             stride > 0 ? stride : -stride) ;
    else
      rep = par ;
    rep->incr() ;
  }
} ;


//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
////                                                                     //
//// struct `Location': The local subscript.                             //
////                                                                     //
//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//
//struct Location : public Coord {
//  Location() {}
//
//  Location(Range x, int glb) {
//    int* sub = subs ;
//    RangeRep* rng ;
//    for(rng = x ; rng->lev() ; rng = rng->ker())
//      glb = rng->ker_glb(glb, sub++) ;
//
//    dim = rng->dim() ;
//    crd = rng->crd(glb) ;
//  }
//
//  int subs [MAX_LEVELS] ;
//
//  // Cosmetic support for ghost regions:
//
//  Location operator<<(const int shift) {
//    Location shifted(*this) ;
//    *(shifted.subs) -= shift ;
//    return shifted ;
//  }
//
//  Location operator>>(const int shift) {
//    Location shifted(*this) ;
//    *(shifted.subs) += shift ;
//    return shifted ;
//  }
//} ;


//inline Location :: Location(const Range& x, int glb) {
//  x.location(this, glb) ;
//}
//
//inline Location Range :: operator()(const int glb) {
//  return Location(rep, glb) ;
//}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// struct `Block'                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

struct Block {
  int count ;

  int glb_bas ;
  int glb_stp ;

  int sub_bas ;
  int sub_stp ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Stride'                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

struct Stride {
public :
  Stride() : itemLen(0) {}

  Stride(Range _parent, const int _itemLen) :
               parent(_parent), itemLen(_itemLen) {}

  Stride ker() const {
    return Stride(parent.ker(), parent.width() * itemLen) ;
  }

  int len() const {return itemLen ; }

  int disp(const int sub) const {
    return parent ? itemLen * parent.disp(sub) : 0 ;
  }

  int disp_step(const int sub_stp) const {
    return parent ? itemLen * parent.step(sub_stp) : 0 ;
  }

  //int offset(const Location& i) const {
  //  return parent ? itemLen * parent.offset(i.subs) : 0 ;
  //}

  int offset(const Location& i) const {
    return parent ? itemLen * parent.offset(i) : 0 ;
  }

  int step(Block& b) const {
    return parent && parent.lev() ? itemLen * parent.step(b.sub_stp) : 0 ;
  }

  //operator int() const {return itemLen ; }

  Range parent ;
  int itemLen ;
} ;

//inline Stride Range :: stride(int* size) const {
//  const int itemLen = *size ;
//  *size *= volume() ;
//  return Stride(*this, itemLen) ;
//}

// Iterator (index) classes

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlkIt': loop state for scanning individual block.            //
//                                                                     //
//                Note that `glb' is computed relative to a            //
//                parent range (``y'') of `x'.                         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlkIt : public Block {
public :
  void init(const Range _x, const int _bas, const int _str) {

    // `_bas', `_str' are alignment parameters of `x' relative to `y'.

    x = _x ;

    bas = _bas ;
    str = _str ;

    glb_stp = str * x.block_step(&sub_stp) ;
  }

  void begin(int* sub, const int ker_glb) {
    glb_bas = bas + str * x.block(ker_glb, &sub_bas, &count) ;

    *sub = sub_bas ;
    glb  = glb_bas ;

    l = 0 ;
  }

  int test() {return l < count ; }

  void next(int* sub) {
    *sub += sub_stp ;
    glb  += glb_stp ;

    l++ ;
  }

  int glb ;

protected :
  Range x ;

  int l ;  // induction variable

  int bas, str ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `LocBlocksIndex': enumerate local blocks of range.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class LocBlocksIndex : public Location, public Block {
public :
  LocBlocksIndex(Range _x) : x(_x), level(_x.lev()) {

    if(level == 0) {
      glb_stp = 1 ;
      d = x ;
    }
    else {
      glb_stp = x.block_step(&sub_stp) ;
      d = x.subker(&ker_glb_bas, &ker_glb_stp) ;

      for(int r = 1 ; r < level ; r++) {
	blk [r].init(d, ker_glb_bas, ker_glb_stp) ;
	d = d.subker(&ker_glb_bas, &ker_glb_stp) ;
      }
    }

    dim = d.dim() ;
  }

  void beginLocBlk() {
    int subker_glb ;
    if(d.local(&subker_glb)) {
      crd = dim.crd() ;

      switch(level) {
        case 0 :
          glb_bas = subker_glb ;
          count   = 1 ;
          looping = 1 ;

          break ;

        case 1 :
          glb_bas = x.block(ker_glb_bas + ker_glb_stp * subker_glb,
                            &sub_bas, &count) ;
          looping = count ;

          *subs   = sub_bas ;
          break ;

        default : {
          looping = 1 ;
          int r = level - 1 ;
          blk [r].begin(subs + r, ker_glb_bas + ker_glb_stp * subker_glb) ;
          scan(r) ;

          *subs   = sub_bas ;
          break ;
        }
      }
    }
    else
      looping = 0 ;
  }

  int test() {return looping ; }

  void nextLocBlk() {
    if(level < 2)
      looping = 0 ;
    else {
      blk [1].next(subs + 1) ;
      scan(1) ;

      *subs   = sub_bas ;
    }
  }

private :

  void scan(int r) {

    // Starting from current position in r'th kernel range,
    // find first non-empty block.

    // 1 <= r < level

    // (Only used for level >= 2.)

    if(blk [r].test()) {
      const int ker_glb = blk [r].glb ;
      if(r == 1) {
        glb_bas = x.block(ker_glb, &sub_bas, &count) ;

        if(count == 0) {
          blk [1].next(subs + 1) ;
          scan(1) ;
        }
      }
      else {
        r-- ;
        blk [r].begin(subs + r, ker_glb) ;
        scan(r) ;
      }
    }
    else {
      r++ ;
      if(r == level)
        looping = 0 ;
      else {
        blk [r].next(subs + r) ;
        scan(r) ;
      }
    }
  }

protected :
  Range x, d ;
  int level ;

  int ker_glb_bas, ker_glb_stp ;  // These relate to innermost, level 0 kernel

  BlkIt blk [MAX_LEVELS] ;  // blk [0] is unused.
  int looping ;

private :
  LocBlocksIndex(const LocBlocksIndex&) {} ;  // No copying.
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Index'.                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Index : public LocBlocksIndex {
public :
  Index(Range _x) : LocBlocksIndex(_x) {}

  void begin() {
    beginLocBlk() ;
    glb = glb_bas ;

    l   = 0 ;
  }

  void next() {
    if(level == 0)
      nextLocBlk() ;
    else {
      *subs += sub_stp ;
      glb   += glb_stp ;

      l++ ;
      if(l == count) {
        nextLocBlk() ;
        glb = glb_bas ;

        l   = 0 ;
      }
    }
  }

  operator int() const {return glb ; }

private :
  int glb ;
  int l ;

} ;


//// Some compilers needs these
//
//// Unfortunately they have the ``wrong'' effect in contexts like
////
////    i + 1.3
//// 
//// ...they lead to the interpretation i + (int) 1.3 instead of
//// (float) (int) i + 1.3.
//
//template<class T>
//inline T operator+(const Index& i, const T n) {
//  return (int) i + n ;
//}
//
//template<class T>
//inline T operator+(const T n, const Index& i) {
//  return n + (int) i ;
//}
//
//inline int operator+(const Index& i, const Index& j) {
//  return (int) i + (int) j ;
//}
//
//template<class T>
//inline T operator-(const Index& i, const T n) {
//  return (int) i - n ;
//}
//
//template<class T>
//inline T operator-(const T n, const Index& i) {
//  return n - (int) i ;
//}
//
//inline int operator-(const Index& i, const Index& j) {
//  return (int) i - (int) j ;
//}
//
//template<class T>
//inline T operator*(const Index& i, const T n) {
//  return (int) i * n ;
//}
//
//template<class T>
//inline T operator*(const T n, const Index& i) {
//  return n * (int) i ;
//}
//
//inline int operator*(const Index& i, const Index& j) {
//  return (int) i * (int) j ;
//}
//
//template<class T>
//inline T operator/(const Index& i, const T n) {
//  return (int) i / n ;
//}
//
//template<class T>
//inline T operator/(const T n, const Index& i) {
//  return n / (int) i ;
//}
//
//inline int operator/(const Index& i, const Index& j) {
//  return (int) i / (int) j ;
//}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `AllBlocksIndex': enumerate all blocks of range.              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class AllBlocksIndex : public LocBlocksIndex {
public :
  AllBlocksIndex(Range _x) : LocBlocksIndex(_x) {}

  void beginAllBlk() {
    switch(level) {
      case 0 :
	crd     = d.bas() ;
	glb_bas = 0 ;

	looping = d.size() ;

        count   = 1 ;
	break ;

      case 1 :
	crd     = d.bas() ;
	ker_glb = ker_glb_bas ;

	l       = 0 ;

	looping = d.size() ;
	if(looping) {
	  glb_bas = x.block(ker_glb, &sub_bas, &count) ;
	  while(count == 0) {
	    crd     += d.str() ;
	    ker_glb += ker_glb_stp ;

	    l++ ;
	    if(l == d.size()) {
	      looping = 0 ;
	      break ;
	    }
	    else
	      glb_bas = x.block(ker_glb, &sub_bas, &count) ;
	  }
	  *subs   = sub_bas ;
	}
	break ;

      default : {
	crd     = d.bas() ;
	ker_glb = ker_glb_bas ;

	l       = 0 ;

        looping = 1 ;
	int r = level - 1 ;
	blk [r].begin(subs + r, ker_glb) ;
	scan(r) ;

	*subs   = sub_bas ;
	break ;
      }
    }
  }

  void nextAllBlk() {
    switch(level) {
      case 0 :
	crd     += d.str() ;
	glb_bas++ ;

	looping = glb_bas < d.size() ;
	break ;

      case 1 :
        do {
	  crd     += d.str() ;
	  ker_glb += ker_glb_stp ;

	  l++ ;

	  if(l == d.size()) {
	    looping = 0 ;
	    break ;
	  }
	  else
	    glb_bas = x.block(ker_glb, &sub_bas, &count) ;
        } while (count == 0) ;

	*subs   = sub_bas ;
	break ;

      default :
	blk [1].next(subs + 1) ;
	scan(1) ;

	*subs   = sub_bas ;
	break ;
    }
  }

private :

  void scan(int r) {

    // Starting from current position in r'th kernel range,
    // find first non-empty block.

    // 1 <= r < level

    // (Only used for level >= 2.)

    if(blk [r].test()) {
      const int ker_glb_r = blk [r].glb ;
      if(r == 1) {
        glb_bas = x.block(ker_glb_r, &sub_bas, &count) ;

        if(count == 0) {
          blk [1].next(subs + 1) ;
          scan(1) ;
        }
      }
      else {
        r-- ;
        blk [r].begin(subs + r, ker_glb_r) ;
        scan(r) ;
      }
    }
    else {
      if(r == level - 1) {
        crd     += d.str() ;
        ker_glb += ker_glb_stp ;

        l++ ;
	if(l == d.size())
	  looping = 0 ;
        else {
          blk [r].begin(subs + r, ker_glb) ;
          scan(r) ;
        }
      }
      else {
        r++ ;
        blk [r].next(subs + r) ;
        scan(r) ;
      }
    }
  }

private :

  int l ;
  int ker_glb ;  // This relates to innermost, level 0 kernel
} ;



// ---------------------------------------------------------------
//
// Things to do:
//
//   Abandon inlining most members of iterators?  The functions are too
//       complex, and confuse compilers (eg, DEC).
//
//   Abandon automatic conversion of index to an arithmetic value (the
//       global subscript).  It doesn't work particularly well---confuses
//       compilers---and in Java it confuses Geoffrey.  Instead need:
//
//         int Range :: idx(const Location& i) ;
//
//   Document changes to `BlockRange', `CyclicRange', `BlockCyclicRange',
//       `StepRange' constructors.
//
//   Move computations in `Stride :: offset' and
//       `Range :: operator()(const int)' to virtual functions on the
//       representation objects, so they can be overridden with
//       efficient implementations (similar to `stride' and `volume').

